from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from wame.engine import Engine

from wame.pipeline import Pipeline
from wame.ui.frame import Frame
from wame.vector import IntVector2

from OpenGL.GL import *

import pygame

class Scene(ABC):
    '''Handles all events and rendering for the engine'''

    def __init__(self, engine:'Engine') -> None:
        '''
        Instantiate a new scene
        
        Parameters
        ----------
        engine : `wame.Engine`
            The engine instance
        '''
        
        self.engine:'Engine' = engine
        '''The engine running the scene'''

        self.screen:pygame.Surface = self.engine.screen
        '''The screen rendering all objects'''

        self.frame:Frame = Frame(engine)
        '''The UI frame responsible for handling all scene UI objects natively - Rendered each frame after `on_render` automatically'''
        self.frame.set_pixel_position((0, 0))
        self.frame.set_pixel_size((self.screen.get_width(), self.screen.get_height()))

        self._first_elapsed:bool = False

    def _check_events(self) -> None:
        for event in pygame.event.get():
            if event.type == pygame.KEYDOWN:
                self.on_key_pressed(event.key, event.mod)
            elif event.type == pygame.KEYUP:
                self.on_key_released(event.key, event.mod)
            elif event.type == pygame.MOUSEBUTTONDOWN:
                mousePosition: IntVector2 = IntVector2.from_tuple(event.pos)

                if event.button in [4, 5]:  # Scrolling shouldn't send a `MOUSEBUTTONDOWN` event
                    continue

                self.on_mouse_pressed(mousePosition, event.button)
            elif event.type == pygame.MOUSEBUTTONUP:
                mousePosition: IntVector2 = IntVector2.from_tuple(event.pos)

                if event.button in [4, 5]:  # Scrolling shouldn't send a `MOUSEBUTTONUP` event
                    continue

                self.on_mouse_released(mousePosition, event.button)
            elif event.type == pygame.MOUSEMOTION:
                mousePosition: IntVector2 = IntVector2.from_tuple(event.pos)

                self.on_mouse_move(mousePosition, IntVector2.from_tuple(event.rel))
            elif event.type == pygame.MOUSEWHEEL:
                mousePosition: IntVector2 = IntVector2.from_tuple(pygame.mouse.get_pos())

                self.on_mouse_wheel_scroll(mousePosition, event.y)
            elif event.type == pygame.QUIT:
                self.engine._running = False

                self.on_quit()
            elif event.type == pygame.WINDOWENTER:
                self.engine._set_fps = self.engine.settings.max_fps
            elif event.type == pygame.WINDOWLEAVE:
                self.engine._set_fps = self.engine.settings.tabbed_fps
    
    def _check_keys(self) -> None:
        keys:pygame.key.ScancodeWrapper = pygame.key.get_pressed()
        mods:int = pygame.key.get_mods()

        for key in range(len(keys)):
            if not keys[key]:
                continue

            self.on_key_pressing(key, mods)
    
    def _cleanup(self) -> None:
        self.on_cleanup()
    
    def _first(self) -> None:
        self.on_first()

        self._first_elapsed = True

    def _render(self) -> None:
        if self.engine._pipeline == Pipeline.PYGAME:
            self.engine.screen.fill(self.engine.background_color.to_tuple())
        elif self.engine._pipeline == Pipeline.OPENGL:
            glClearColor(self.engine.background_color.nr, self.engine.background_color.ng, self.engine.background_color.nb, self.engine.background_color.a)

        self.on_render()
        self.frame.ask_render()

        pygame.display.flip()
        self.engine._deltaTime = self.engine._clock.tick(self.engine._set_fps) / 1000.0

    def _update(self) -> None:
        self.on_update()
    
    def on_cleanup(self) -> None:
        '''
        Code below should be executed when the scene is being switched/cleaned up

        Example
        -------
        ```python
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_cleanup(self) -> None:
                ... # Terminate background threads, save data, etc.
        ```
        '''

        ...

    def on_first(self) -> None:
        '''
        Code below should be executed when the scene is about to start rendering

        Example
        -------
        ```python
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_first(self) -> None:
                ... # Start game timers, etc.
        ```
        '''

        ...

    def on_key_pressed(self, key:int, mods:int) -> None:
        '''
        Code below should be executed when a key is pressed

        Example
        -------
        ```python
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_key_pressed(self, key:int, mods:int) -> None:
                ... # Pause game, display UI, etc.
        ```
        '''
        
        ...
    
    def on_key_pressing(self, key:int, mods:int) -> None:
        '''
        Code below should be executed when a key is being pressed

        Example
        -------
        ```python
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_key_pressing(self, key:int, mods:int) -> None:
                ... # Move forward, honk horn, etc.
        ```
        '''
        
        ...
    
    def on_key_released(self, key:int, mods:int) -> None:
        '''
        Code below should be executed when a key is released

        Example
        -------
        ```python
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_key_released(self, key:int, mods:int) -> None:
                ... # Stop moving forward, etc.
        ```
        '''
        
        ...
    
    def on_mouse_move(self, mouse_position:IntVector2, relative:IntVector2) -> None:
        '''
        Code below should be executed when the mouse moves

        Example
        -------
        ```python
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_mouse_move(self, mouse_position:wame.IntVector2, relative:wame.IntVector2) -> None:
                print(f"Mouse was moved {relative} amount @ {mouse_position}")
        ```
        '''
        
        ...
    
    def on_mouse_pressed(self, mouse_position:IntVector2, button:int) -> None:
        '''
        Code below should be executed when a mouse button was pressed

        Example
        -------
        ```python
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_mouse_pressed(self, mouse_position:wame.IntVector2, button:int) -> None:
                ... # Start shooting, rotate character, etc.
        ```
        '''
        
        ...
    
    def on_mouse_released(self, mouse_position:IntVector2, button:int) -> None:
        '''
        Code below should be executed when a mouse button was released

        Example
        -------
        ```python
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_mouse_released(self, mouse_position:wame.IntVector2, button:int) -> None:
                ... # Shoot arrow, stop shooting, etc.
        ```
        '''
        
        ...
    
    def on_mouse_wheel_scroll(self, mouse_position:IntVector2, amount:int) -> None:
        '''
        Code below should be executed when the scroll wheel moves

        Example
        -------
        ```python
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_mouse_wheel_scroll(self, mouse_position:wame.IntVector2, amount:int) -> None:
                if amount > 0:
                    print(f"Scroll wheel moved up @ {mouse_position}!")
                else:
                    print(f"Scroll wheel moved down @ {mouse_position}!")
        ```
        '''

        ...

    @abstractmethod
    def on_render(self) -> None:
        '''
        Code below should be executed every frame to render all objects after being updated
        
        Example
        -------
        ```python
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_render(self) -> None:
                ... # Render text, objects, etc.
        ```
        '''

        ...

    @abstractmethod
    def on_update(self) -> None:
        '''
        Code below should be executed every frame before objects are rendered to provide updates to instance states

        Example
        -------
        ```python
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_update(self) -> None:
                ... # Update positions, text, etc.
        ```
        '''
        
        ...
    
    def on_quit(self) -> None:
        '''
        Code below should be executed when the engine quits

        Example
        -------
        ```python
        class MyScene(wame.Scene):
            def __init__(self, engine) -> None:
                super().__init__(engine)
            
            def on_quit(self) -> None:
                ... # Save data, cleanup objects, etc.
        ```
        '''
        
        ...


from __future__ import annotations

from wame.vector import IntVector2

import pygame
import wame

class Renderable:
    '''Base UI object that all `wame` UI objects inherit'''

    def __init__(self, engine:'wame.Engine') -> None:
        '''
        Instantiate a renderable UI object
        
        Parameters
        ----------
        engine : `wame.Engine`
            The engine to hook this UI object to
        '''
        
        self._engine:'wame.Engine' = engine

        self.enabled:bool = True
        '''If this renderable should render on the screen'''

        self.position:IntVector2 = None
        '''The absolute X, Y position (pixel position) of this object'''

        self.size:IntVector2 = None
        '''The absolute X, Y size (pixel amount) of this object'''

        self._children:list[Renderable] = []

    def add_child(self, child:'Renderable') -> None:
        '''
        Add a subordinate child object to this UI object
        
        Parameters
        ----------
        child : `wame.ui.renderable.Renderable`
            The child renderable object
        '''

        if not isinstance(child, Renderable):
            error:str = "Child object must be an instance of `wame.ui.renderable.Renderable`"
            raise ValueError(error)
    
        self._children.append(child)

    def ask_render(self) -> None:
        '''Request to render if this renderable is enabled'''
        
        if not self.enabled:
            return
        
        self.render()

    @property
    def rect(self) -> pygame.Rect:
        return pygame.Rect(self.position.to_tuple(), self.size.to_tuple())

    def remove_child(self, child:'Renderable') -> None:
        '''
        Remove a subordinate child object from this UI object
        
        Parameters
        ----------
        child : `wame.ui.renderable.Renderable`
            The child renderable object
        '''

        if not isinstance(child, Renderable):
            error:str = "Child object must be an instance of `wame.ui.renderable.Renderable`"
            raise ValueError(error)
        
        try:
            self._children.remove(child)
        except ValueError:
            error:str = "Child object was not found in the parent frame"
            raise ValueError(error)

    def render(self) -> None:
        '''
        Render this object to the screen
        
        Note
        ----
        This should only be called by the `~ask_render` method, as this ignores the objects `enabled` attribute.'''
        
        for child in self._children:
            child.ask_render()


from __future__ import annotations

class Pipeline:
    '''Engine Rendering Pipeline'''

    PYGAME:int = 0
    '''
    PyGame will render all elements/objects
    '''

    OPENGL:int = 1
    '''
    OpenGL will render all elements/objects
    '''


'''
## UI Module - Wame Engine
Contains all necessary objects and functionality for rendering user interfaces
'''

from wame.ui.button import Button
from wame.ui.frame import Frame
from wame.ui.image import Image
from wame.ui.input import CheckboxInput, TextInput
from wame.ui.renderable import Renderable
from wame.ui.text import Text


from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from wame.engine import Engine

from wame.pipeline import Pipeline

import pygame

class Settings:
    '''Engine Global Settings'''

    def __init__(self, data:dict[str, int], engine:'Engine') -> None:
        '''
        Initialize the engine settings
        
        Note
        ----
        This should only be instantiated by the engine internally.
        Avoid calling this yourself.
        
        Parameters
        ----------
        data : `dict[str, Any]`
            The raw settings contents from the known settings persistence file
        engine : `wame.engine.Engine`
            The engine using these settings
        '''
        
        self._engine:'Engine' = engine

        self._antialiasing:int = data["antialiasing"] if "antialiasing" in data else 0

        self._max_fps:int = data["max_fps"] if "max_fps" in data else 0

        self._tabbed_fps:int = data["tabbed_fps"] if "tabbed_fps" in data else 30

        self._vsync:int = data["vsync"] if "vsync" in data else 0

    @property
    def antialiasing(self) -> bool:
        '''Graphics technique used to remove jagged edges'''

        return bool(self._antialiasing)
    
    @antialiasing.setter
    def antialiasing(self, value:bool) -> None:
        if not isinstance(value, bool):
            error:str = "Antialiasing setting must be a boolean"
            raise ValueError(error)
        
        self._antialiasing = int(value)
    
    @property
    def max_fps(self) -> int:
        '''The maximum framerate that the engine will render scenes at'''

        return self._max_fps
    
    @max_fps.setter
    def max_fps(self, value:int) -> None:
        if not isinstance(value, int):
            error:str = "Max FPS value must be an integer"
            raise ValueError(error)
        
        if value < 0:
            error:str = "Max FPS value must be 0 or above"
            raise ValueError(error)
        
        self._max_fps = value
        self._engine._set_fps = self._max_fps
    
    @property
    def tabbed_fps(self) -> int:
        '''The maximum framerate that the engine will render scenes at when the cursor is outside the application'''

        return self._tabbed_fps

    @tabbed_fps.setter
    def tabbed_fps(self, value:int) -> None:
        if not isinstance(value, int):
            error:str = "Tabbed FPS value must be an integer"
            raise ValueError(error)
        
        if value < 0:
            error:str = "Tabbed FPS value must be 0 or above"
            raise ValueError(error)
        
        self._tabbed_fps = value
    
    @property
    def vsync(self) -> bool:
        '''Syncs the graphics card's framerate with the refresh rate of the display device'''

        return bool(self._vsync)
    
    @vsync.setter
    def vsync(self, value:bool) -> None:
        if not isinstance(value, bool):
            error:str = "VSync value must be a boolean"
            raise ValueError(error)
        
        self._vsync = int(value)

        if self._engine._pipeline == Pipeline.PYGAME:
            self.screen = pygame.display.set_mode(self._engine._size.to_tuple(), pygame.HWSURFACE | pygame.DOUBLEBUF, display=self._engine._display, vsync=self._vsync)
        else:
            self.screen = pygame.display.set_mode(self._engine._size.to_tuple(), pygame.HWSURFACE | pygame.DOUBLEBUF | pygame.OPENGL, display=self._engine._display, vsync=self._vsync)

    def export(self) -> dict[str, int]:
        '''
        Export the class instance to a raw, storable type
        
        Returns
        -------
        rawSettings : `dict[str, int]`
            The raw class instance data
        '''
        
        return {
            "antialiasing": self._antialiasing,
            "max_fps": self._max_fps,
            "tabbed_fps": self._tabbed_fps,
            "vsync":self._vsync
        }


from __future__ import annotations

from wame.color.rgb import ColorRGBA
from wame.vector.xy import FloatVector2, IntVector2
from wame.ui.renderable import Renderable

from OpenGL.GLU import *
from OpenGL.GL import *

import pygame
import wame

# TODO: Border Radius

class Frame(Renderable):
    '''UI Container'''

    def __init__(self, parent:'Frame', *, color:ColorRGBA=None, y_flipped:bool=False) -> None:
        '''
        Create a UI frame
        
        Parameters
        ----------
        parent : `wame.ui.frame.Frame`
            The parent of this frame
        color : `wame.color.rgb.ColorRGBA`
            The background color of the frame
        y_flipped : `bool`
            If it should be rendered with the Y-axis flipped - May be necessary depending on your OpenGL setup
        
        Note
        ----
        Scenes already natively contain a UI frame. Unless you want to make a sub-frame to encapsulate other child renderables, using the `Scene`'s `frame` attribute should be sufficient
        '''

        super().__init__(parent if isinstance(parent, wame.Engine) else parent._engine)

        if isinstance(parent, Frame):
            parent.add_child(self)
            
            self._parent = parent
        else: # If natively set to the engine, this is the scene's native frame (no parent)
            self._parent = None

        self._children:list[Renderable] = []

        self._color:ColorRGBA = (color if isinstance(color, ColorRGBA) else ColorRGBA.from_tuple(color)) if color else None

        self._border_color:ColorRGBA = None
        self._border_width:int = None

        self._flipped:bool = y_flipped

    def render(self) -> None:
        '''
        Render this frame and it's children to the screen
        '''
        
        if not self.position or not self.size:
            error:str = "The frame must have its size and position set before rendering"
            raise ValueError(error)
    
        if self._color:
            if self._engine._pipeline == wame.Pipeline.PYGAME:
                pygame.draw.rect(
                    self._engine.screen, self._color.to_tuple(), self.rect
                )
            elif self._engine._pipeline == wame.Pipeline.OPENGL:
                posY:int = self._engine.screen.get_height() - (self.position.y + self.size.y)
                posWidth:int = self.position.x + self.size.x
                posHeight:int = self.position.y + self.size.y

                glMatrixMode(GL_PROJECTION)
                glLoadIdentity()

                gluOrtho2D(0, self._engine.screen.get_width(), 0, self._engine.screen.get_height())
                glMatrixMode(GL_MODELVIEW)
                glLoadIdentity()

                glPushMatrix()

                glDisable(GL_LIGHTING)
                glDisable(GL_TEXTURE_2D)

                glColor4f(self._color.r / 255, self._color.g / 255, self._color.b / 255, self._color.a)

                glBegin(GL_QUADS)
                if self._flipped:
                    glVertex2f(self.position.x, posHeight)
                    glVertex2f(posWidth, posHeight)
                    glVertex2f(posWidth, posY)
                    glVertex2f(self.position.x, posY)
                else:
                    glVertex2f(self.position.x, posY)
                    glVertex2f(posWidth, posY)
                    glVertex2f(posWidth, posHeight)
                    glVertex2f(self.position.x, posHeight)
                glEnd()

                glPopMatrix()
        
        if self._border_color and self._border_width >= 1:
            # Lines are straight, no point in antialiasing them

            if self._engine._pipeline == wame.Pipeline.PYGAME:
                for index in range(self._border_width):
                    pygame.draw.lines(self._engine.screen, self._border_color.to_tuple(), True, [
                        (self.position.x + index, self.position.y + index), (self.position.x + self.size.x - index, self.position.y + index),
                        (self.position.x + self.size.x - index, self.position.y + self.size.y - index), (self.position.x + index, self.position.y + self.size.y - index)
                    ])
            elif self._engine._pipeline == wame.Pipeline.OPENGL:
                ...

        for child in self._children:
            child.ask_render()
    
    def set_border(self, color:ColorRGBA, width:int) -> None:
        self._border_color = color if isinstance(color, ColorRGBA) else ColorRGBA.from_tuple(color)
        self._border_width = width

    def set_color(self, color:ColorRGBA) -> None:
        self._color = color if isinstance(color, ColorRGBA) else ColorRGBA.from_tuple(color)

    def set_pixel_position(self, position:IntVector2) -> None:
        position = position if isinstance(position, IntVector2) else IntVector2.from_tuple(position)

        if self._parent:
            position.x += self._parent.position.x
            position.y += self._parent.position.y
        
        self.position = position

    def set_pixel_size(self, size:IntVector2) -> None:
        self.size = size if isinstance(size, IntVector2) else IntVector2.from_tuple(size)

    def set_scaled_position(self, position:FloatVector2) -> None:
        position = position if isinstance(position, FloatVector2) else FloatVector2.from_tuple(position)

        if position.x > 1 or position.x < 0 or position.y > 1 or position.y < 0:
            error:str = "Scaled position X, Y values must be between 0 and 1"
            raise ValueError(error)
        
        newPosition:IntVector2 = IntVector2(0, 0)

        if not self._parent:
            newPosition.x = int(self._engine.screen.get_width() * position.x)
            newPosition.y = int(self._engine.screen.get_height() * position.y)
        else:
            newPosition.x = int(self._parent.position.x + (self._parent.size.x * position.x))
            newPosition.y = int(self._parent.position.y + (self._parent.size.y * position.y))

        self.position = newPosition
    
    def set_scaled_size(self, size:IntVector2) -> None:
        size = size if isinstance(size, FloatVector2) else FloatVector2.from_tuple(size)

        if size.x > 1 or size.x < 0 or size.y > 1 or size.y < 0:
            error:str = "Scaled size X, Y values must be between 0 and 1"
            raise ValueError(error)
    
        newSize:IntVector2 = IntVector2(0, 0)

        if not self._parent:
            newSize.x = int(self._engine.screen.get_width() * size.x)
            newSize.y = int(self._engine.screen.get_height() * size.y)
        else:
            newSize.x = int(self._parent.size.x * size.x)
            newSize.y = int(self._parent.size.y * size.y)

        self.size = newSize


from __future__ import annotations

from typing import NoReturn

class ColorRGB:
    '''RGB Color Object'''

    def __init__(self, r:int, g:int, b:int) -> None:
        '''
        Create a RGB color
        
        Parameters
        ----------
        r : `int`
            The R (red) value (0-255)
        g : `int`
            The G (green) value (0-255)
        b : `int`
            The B (blue) value (0-255)
        '''

        def color_not_int(color:str) -> NoReturn:
            error:str = f"RGB color value {color} must be an integer"
            raise ValueError(error)
    
        def color_invalid_range(color:str, value:int) -> NoReturn:
            error:str = f"RGB color value {color} value must be between 0-255, not {value}"
            raise ValueError(error)

        if not isinstance(r, int):
            color_not_int('R')
        
        if not isinstance(g, int):
            color_not_int('G')
        
        if not isinstance(b, int):
            color_not_int('B')
        
        if r > 255 or r < 0:
            color_invalid_range('R', r)
        
        if g > 255 or g < 0:
            color_invalid_range('G', g)
        
        if b > 255 or b < 0:
            color_invalid_range('B', b)

        self.r:int = r
        self.g:int = g
        self.b:int = b

        self.nr:float = r / 255
        self.ng:float = g / 255
        self.nb:float = b / 255

    def __str__(self) -> str:
        return f"R: {self.r}, G: {self.g}, B: {self.b}"

    @classmethod
    def from_tuple(cls, rgb:tuple[int, int, int]) -> ColorRGB:
        '''
        Create a RGB color from a tuple
        
        Parameters
        ----------
        rgb : `tuple[int, int, int]`
            The R, G, and B values
        
        Returns
        -------
        color : `wame.color.rgb.ColorRGB`
            New `ColorRGB` object from the tuple
        '''

        return ColorRGB(rgb[0], rgb[1], rgb[2])
    
    def normalized(self) -> tuple[float, float, float]:
        '''
        Normalize this color into RGB values `0-1`, not `0-255`
        
        Returns
        -------
        rgb : `tuple[float, float, float]`
            The normalized RGB values
        '''

        return (self.nr, self.ng, self.nb)

    def to_tuple(self) -> tuple[int, int, int]:
        '''
        Convert this object to a tuple
        
        Returns
        -------
        color : `tuple[int, int, int]`
            The R, G, and B values in a tuple
        '''

        return (self.r, self.g, self.b)

class ColorRGBA(ColorRGB):
    '''RGBA Color Object'''

    def __init__(self, r:int, g:int, b:int, a:float) -> None:
        '''
        Create a RGB color
        
        Parameters
        ----------
        r : `int`
            The R (red) value (0-255)
        g : `int`
            The G (green) value (0-255)
        b : `int`
            The B (blue) value (0-255)
        a : `int`
            The A (alpha) value (0-1)
        '''

        super().__init__(r, g, b)
        
        if not isinstance(a, (int, float)):
            error:str = f"RGBA color value A must be either an integer or float"
            raise ValueError(error)

        if a > 1 or a < 0:
            error:str = f"RGBA color value A value must be between 0-1, not {a}"
            raise ValueError(error)

        self.a:float = a
    
    def __str__(self) -> str:
        return f"R: {self.r}, G: {self.g}, B: {self.b}, A: {self.a}"

    @classmethod
    def from_tuple(cls, rgba:tuple[int, int, int, float]) -> ColorRGBA:
        '''
        Create a RGBA color from a tuple
        
        Parameters
        ----------
        rgba : `tuple[int, int, int, float]`
            The R, G, B, and (optionally) A values
        
        Returns
        -------
        color : `wame.color.rgb.ColorRGBA`
            New `ColorRGBA` object from the tuple - If `A` is omitted, default is `1.0`
        '''

        try:
            return ColorRGBA(rgba[0], rgba[1], rgba[2], rgba[3])
        except IndexError:
            return ColorRGBA(rgba[0], rgba[1], rgba[2], 1.0)

    def normalized(self) -> tuple[float, float, float, float]:
        '''
        Normalize this color into RGBA values `0-1`, not `0-255`
        
        Returns
        -------
        rgba : `tuple[float, float, float, float]`
            The normalized RGBA values
        '''

        return (self.nr, self.ng, self.nb, self.a)

    def to_tuple(self) -> tuple[int, int, int, float]:
        '''
        Convert this object to a tuple
        
        Returns
        -------
        color : `tuple[int, int, int, float]`
            The R, G, B, and A values in a tuple
        '''

        return (self.r, self.g, self.b, self.a)


from __future__ import annotations

from wame.color.rgb import ColorRGBA
from wame.vector import IntVector2
from wame.settings import Settings
from wame.pipeline import Pipeline
from wame.scene import Scene

import importlib
import pygame
import json
import ast
import os

pygame.init()
pygame.font.init()

class Engine:
    '''Game Engine'''
    
    _previously_instantiated:bool = False

    def __init__(self, name:str, pipeline:Pipeline, *, size:IntVector2=IntVector2(0, 0), display:int=0, icon_filepath:str=None) -> None:
        '''
        Instantiates a game engine that handles all backend code for running games
        
        Parameters
        ----------
        name : `str`
            The name of the engine window
        pipeline : `wame.Pipeline`
            The pipeline library the engine should use
        size : `wame.vector.IntVector2`
            The X and Y sizes for the game window
        display : `int`
            The index of the display/monitor for the rendering screen
        icon_filepath : `str`
            The filepath of the game window icon
        '''

        if Engine._previously_instantiated:
            error:str = "Only one instance of `wame.engine.Engine` is supported during runtime"
            raise RuntimeError(error)
        
        Engine._previously_instantiated = True

        self._name:str = name

        self.screen:pygame.Surface = None
        '''The screen that the engine is rendering to'''
        self._clock:pygame.time.Clock = pygame.time.Clock()
        self._deltaTime:float = 0.001
        self._running:bool = False

        if not os.path.exists("settings.json"):
            with open("settings.json", 'w') as file:
                file.write("{}")
        
        with open("settings.json") as file:
            self.settings:Settings = Settings(json.load(file), self)
            '''The settings that the engine renders/runs the game with'''

        self.scene:Scene = None
        '''The scene that is currently active - `None` if no scene is running'''
        self.scenes:dict[str, Scene] = {}
        '''The key-value pairs of scenes and their names (name: scene)'''

        self._set_fps:int = self.settings.max_fps
        self._background_color:ColorRGBA = ColorRGBA(0, 0, 0, 1.0)

        self._size:IntVector2 = size

        self._mouse_visibility:bool = True
        self._mouse_grabbed:bool = False
        
        self._pipeline:Pipeline = pipeline

        if pipeline not in [Pipeline.PYGAME, Pipeline.OPENGL]:
            error:str = "Sorry, the requested pipeline is not supported"
            raise RuntimeError(error)
        
        self._display:int = display
        self.set_pipeline(pipeline)

        if icon_filepath:
            pygame.display.set_icon(pygame.image.load(icon_filepath.replace('\\', '/')))

        pygame.display.set_caption(self._name)

        pygame.mouse.set_visible(self._mouse_visibility)
        pygame.event.set_grab(self._mouse_grabbed)

    def _cleanup(self) -> None:
        with open("settings.json", 'w') as file:
            json.dump(self.settings.export(), file, indent=4)

    def _mainloop(self) -> None:
        if not self.scene:
            error:str = "A starting scene must be defined before the engine can start. Register a scene with any engine.register_scene ... and set the scene using engine.set_scene()"
            raise RuntimeError(error)

        self._running = True

        while self._running:
            self.scene._check_events()
            self.scene._check_keys()
            self.scene._update()

            self.scene._render()
        
        self.scene._cleanup()
        self._cleanup()

    @property
    def background_color(self) -> ColorRGBA:
        '''The background/screen color of the engine'''

        return self._background_color

    @property
    def delta_time(self) -> float:
        '''Time since the last frame was rendered'''

        return self._deltaTime

    @property
    def fps(self) -> float:
        '''Frames per second of the engine'''

        return self._clock.get_fps()

    @property
    def mouse_locked(self) -> bool:
        '''If the mouse is locked to the engine window'''

        return self._mouse_grabbed

    @property
    def mouse_visible(self) -> bool:
        '''If the mouse is visible'''

        return self._mouse_visibility

    @property
    def pipeline(self) -> Pipeline:
        '''The current pipeline of the engine'''

        return self._pipeline

    def quit(self) -> None:
        '''
        Stops the engine and cleans up
        '''
        
        self._running = False
        self.scene.on_quit()

    def register_scene(self, name:str, scene:Scene, overwrite:bool=False) -> None:
        '''
        Register a scene to the engine
        
        Parameters
        ----------
        name : `str`
            The unique name used to lookup and manipulate this scene
        scene : `wame.Scene`
            The scene to register
        overwrite : `bool`
            If the unique name is already used, overwrite it, else throw an error - Default `False`
        
        Raises
        ------
        `wame.UniqueNameAlreadyExists`
            If the unique name already exists and overwriting is not enabled
        '''
        
        if not overwrite:
            if name in self.scenes:
                error:str = f"Scene name \"{name}\" already in use"
                raise RuntimeError(error)

        self.scenes[name] = scene

    def register_scenes(self, scenes:dict[str, Scene], overwrite:bool=False) -> None:
        '''
        Register a set of scenes to the engine
        
        Parameters
        ----------
        scenes : `dict[str, wame.Scene]`
            The name-scene pairs to register
        overwrite : `bool`
            If any unique name is already used, overwrite it, else throw an error - Default `False`
        
        Raises
        ------
        `wame.UniqueNameAlreadyExists`
            If any unique name already exists and overwriting is not enabled
        '''

        for name, scene in scenes.items():
            self.register_scene(name, scene, overwrite)
    
    def register_scenes_from_folder(self, folder:str, overwrite:bool=False) -> None:
        '''
        Register all Scene objects within all files in a folder to the engine
        
        Warning
        -------
        If you plan on bundling this game into an executable file:
        - Continue to use this method, but also include the raw scene program files in the folder provided as well as the .exe file OR
        - Manually register each scene
        #### This is because to bundle Python into executable files, there must be a direct reference to dependencies. Hotloading scenes has no direct reference.

        Note
        ----
        Folder must be in the same directory as your project.
        The engine will only walk through the files in this folder, not any subdirectories.
        All unique scene names will be generated from the Scene subclass names themselves:
        ```python
        class MyScene(wame.Scene):
            ...
        ```
        Will generate unique name "My" and can be used to set the scene later on

        ```python
        class MainMenuScene(wame.Scene):
            ...
        ```
        Will generate unique name "MainMenu" and can be used to set the scene later on

        And so forth...
        
        Parameters
        ----------
        folder : `str`
            The folder to register scenes from
        overwrite : `bool`
            If any unique name is already used, overwrite it, else throw an error - Default `False`
        
        Raises
        ------
        `wame.SceneFolderNotFound`
            If the folder path provided does not exist or if the folder path does not direct to a folder
        `wame.UniqueNameAlreadyExists`
            If any unique name already exists and overwriting is not enabled
        '''
        
        if not os.path.exists(folder):
            error:str = f"Folder \"{folder}\" could not be found"
            raise RuntimeError(error)
        
        if not os.path.isdir(folder):
            error:str = f"Item with name \"{folder}\" is not a folder/directory"
            raise RuntimeError(error)
        
        for filename in os.listdir(folder):
            if not filename.endswith(".py"):
                continue

            with open(f"{folder}/{filename}") as file:
                contents:str = file.read()
            
            tree:ast.Module = ast.parse(contents)
            classes:list[ast.ClassDef] = [node for node in ast.walk(tree) if isinstance(node, ast.ClassDef)]

            for fileClass in classes:
                endIndex:int = fileClass.name.find("Scene")

                if endIndex < 0:
                    continue

                sceneName:str = fileClass.name[0:endIndex]
                
                module = importlib.import_module(f"{folder}.{filename[:-3]}")
                sceneObject:Scene = getattr(module, fileClass.name)

                self.register_scene(sceneName, sceneObject, overwrite)

    @property
    def running(self) -> bool:
        '''If the engine is still running the game loop'''

        return self._running

    def set_background(self, color:ColorRGBA=ColorRGBA(0, 0, 0, 1.0)) -> None:
        '''
        Set the background color of the engine rendering scene
        
        Parameters
        ----------
        color : `wame.color.rgb.ColorRGBA`
            The background color to apply to all scenes - Default `ColorRGBA(0, 0, 0, 1.0)`
        '''

        if isinstance(color, tuple):
            color = ColorRGBA.from_tuple(color)

        self._background_color = color

    def set_mouse_visible(self, state:bool=True) -> None:
        '''
        Set if the mouse should be visible or hidden
        
        Parameters
        ----------
        state : `bool`
            If the mouse should be visible or hidden - Default `True`
        '''

        self._mouse_visibility = state
        pygame.mouse.set_visible(state)

    def set_mouse_locked(self, state:bool=False) -> None:
        '''
        Set if the mouse should be immovable
        
        Parameters
        ----------
        state : `bool`
            If the mouse should be locked or not - Default `False`
        '''

        self._mouse_grabbed = state
        pygame.event.set_grab(state)

    def set_pipeline(self, pipeline:Pipeline) -> None:
        '''
        Set the rendering pipeline that the engine should use
        
        Parameters
        ----------
        pipeline : `wame.Pipeline`
            The rendering pipeline to switch to
        '''

        if self.scene and self.scene._first_elapsed:
            error:str = "Switching the rendering pipeline during the game loop is not supported"
            raise RuntimeError(error)

        self._pipeline = pipeline

        if pipeline == Pipeline.PYGAME:
            self.screen = pygame.display.set_mode(self._size.to_tuple(), pygame.HWSURFACE | pygame.DOUBLEBUF, display=self._display, vsync=self.settings.vsync)
        else:
            self.screen = pygame.display.set_mode(self._size.to_tuple(), pygame.HWSURFACE | pygame.DOUBLEBUF | pygame.OPENGL, display=self._display, vsync=self.settings.vsync)

    def set_scene(self, name:str, *args) -> None:
        '''
        Switch the engine to another scene and clean up the previous (if any)
        
        Parameters
        ----------
        name : `str`
            The unique name of the scene to switch to (must be previously registered)
        args : `Any`
            Any data you wish to pass to the scene instance
        
        Raises
        ------
        `wame.UniqueNameNotFound`
            If the name does not exist
        '''
        
        if name not in self.scenes:
            error:str = f"Scene with name \"{name}\" was not registered/found"
            raise RuntimeError(error)
        
        if isinstance(self.scene, self.scenes[name]):
            error:str = f"Scene with name \"{name}\" is already set as the active scene"
            raise RuntimeError(error)
        
        if self.scene is not None:
            self.scene._cleanup()
            del self.scene

        self.scene = self.scenes[name](self, *args)
        self.scene._first()

    def start(self) -> None:
        '''
        Starts the engine

        Warning
        -------
        This is a blocking call. No code below will execute until the engine has stopped running.

        Raises
        ------
        `wame.SceneNotSet`
            If the engine is started without a scene registered and set
        '''
        
        self._mainloop()


from __future__ import annotations

from typing import Union

import numpy as np
import math

class FloatVector3:
    '''Vector with 3 Float Values: X, Y, and Z'''

    def __init__(self, x:float, y:float, z:float) -> None:
        '''
        Instantiate a new Vector with float X, Y, and Z values
        
        Parameters
        ----------
        x : `float`
            The X value
        y : `float`
            The Y value
        z : `float`
            The Z value
        
        Raises
        ------
        `ValueError`
            If any provided arguments are not floats
        '''
        
        if isinstance(x, int):
            x = float(x)
        
        if isinstance(y, int):
            y = float(y)
        
        if isinstance(z, int):
            z = float(z)

        self.x:float = x
        self.y:float = y
        self.z:float = z
    
    def __add__(self, other:Union[FloatVector3, IntVector3, np.ndarray]) -> FloatVector3:
        if isinstance(other, FloatVector3) or isinstance(other, IntVector3):
            return FloatVector3(self.x + other.x, self.y + other.y, self.z + other.z)
        
        if isinstance(other, np.ndarray):
            return FloatVector3(self.x + other[0], self.y + other[1], self.z + other[2])
        
        raise TypeError(f"Unsupported addition with type {type(other)}")
    
    def __eq__(self, other:Union[FloatVector3, IntVector3, np.ndarray]) -> bool:
        if isinstance(other, FloatVector3) or isinstance(other, IntVector3):
            return self.to_tuple() == other.to_tuple()
        
        if isinstance(other, np.ndarray):
            return self.to_tuple() == (other[0], other[1], other[2])

    def __hash__(self) -> int:
        return hash(self.to_tuple())

    def __sub__(self, other:Union[FloatVector3, IntVector3, np.ndarray]) -> FloatVector3:
        if isinstance(other, FloatVector3) or isinstance(other, IntVector3):
            return FloatVector3(self.x - other.x, self.y - other.y, self.z - other.z)
        
        if isinstance(other, np.ndarray):
            return FloatVector3(self.x - other[0], self.y - other[1], self.z - other[2])
        
        raise TypeError(f"Unsupported subtraction with type {type(other)}")

    def __repr__(self) -> str:
        return f"X: {self.x}, Y: {self.y}, Z: {self.z}"
    
    @classmethod
    def from_tuple(cls, xyz:tuple[float, float, float]) -> FloatVector3:
        '''
        Instantiate a new Vector from a tuple with float X, Y, and Z values
        
        Parameters
        ----------
        xyz : `tuple[float, float, float]`
            The tuple with the X, Y, and Z values
        
        Raises
        ------
        `ValueError`
            If the provided items in the tuple are not floats
        '''
        
        return cls(xyz[0], xyz[1], xyz[2])
    
    def to_numpy(self) -> np.ndarray[np.int32]:
        '''
        Converts this instance of `FloatVector3` into a numpy array
        
        Returns
        -------
        array : `numpy.ndarray[numpy.float32]`
            Converted x, y, and z values
        '''

        return np.array([self.x, self.y, self.z], dtype=np.float32)

    def to_tuple(self) -> tuple[float, float, float]:
        '''
        Converts this instance of `FloatVector3` into a `tuple`
        
        Returns
        -------
        vector : `tuple[float, float, float]`
            Converted x, y, and z values
        '''

        return (self.x, self.y, self.z)

class IntVector3:
    '''Vector with 3 Integer Values: X, Y, and Z'''

    def __init__(self, x:int, y:int, z:int) -> None:
        '''
        Instantiate a new Vector with integer X, Y, and Z values
        
        Parameters
        ----------
        x : `int`
            The X value
        y : `int`
            The Y value
        z : `int`
            The Z value
        
        Raises
        ------
        `ValueError`
            If any provided arguments are not integers
        '''
        
        if not isinstance(x, int):
            error:str = f"Value of X ({x}) is not an integer"
            raise ValueError(error)
    
        if not isinstance(y, int):
            error:str = f"Value of Y ({y}) is not an integer"
            raise ValueError(error)
    
        if not isinstance(z, int):
            error:str = f"Value of Z ({z}) is not an integer"
            raise ValueError(error)

        self.x:int = x
        self.y:int = y
        self.z:int = z
    
    def __add__(self, other:Union[IntVector3, np.ndarray]) -> IntVector3:
        if isinstance(other, IntVector3):
            return IntVector3(self.x + other.x, self.y + other.y, self.z + other.z)
        
        if isinstance(other, np.ndarray):
            return IntVector3(self.x + other[0], self.y + other[1], self.z + other[2])
        
        raise TypeError(f"Unsupported addition with type {type(other)}")

    def __hash__(self) -> int:
        return hash(self.to_tuple())
    
    def __eq__(self, other:Union[FloatVector3, IntVector3, np.ndarray]) -> bool:
        if isinstance(other, FloatVector3) or isinstance(other, IntVector3):
            return self.to_tuple() == other.to_tuple()
        
        if isinstance(other, np.ndarray):
            return self.to_tuple() == (other[0], other[1], other[2])

    def __sub__(self, other:Union[IntVector3, np.ndarray]) -> IntVector3:
        if isinstance(other, IntVector3):
            return IntVector3(self.x - other.x, self.y - other.y, self.z - other.z)
        
        if isinstance(other, np.ndarray):
            return IntVector3(self.x - other[0], self.y - other[1], self.z - other[2])
        
        raise TypeError(f"Unsupported subtraction with type {type(other)}")

    def __repr__(self) -> str:
        return f"X: {self.x}, Y: {self.y}, Z: {self.z}"
    
    def cross(self, vector:Union[IntVector3, 'FloatVector3']) -> Union[IntVector3, 'FloatVector3']:
        '''
        Calculate the cross product of two vectors
        
        Parameters
        ----------
        vector : `wame.IntVector3, wame.FloatVector3`
            The other vector to cross with this vector
        
        Returns
        -------
        cross : `wame.IntVector3, wame.FloatVector3`
            The cross product - Will be `wame.IntVector3` if `vector` is a `wame.IntVector3`, else `wame.FloatVector3`
        '''

        result:tuple[float, int, float, int, float, int] = (
            (self.y * vector.z) - (self.z * vector.y),
            (self.z * vector.x) - (self.x * vector.z),
            (self.x * vector.y) - (self.y * vector.x)
        )

        if isinstance(vector, IntVector3):
            if isinstance(result[0], int) and isinstance(result[1], int) and isinstance(result[2], int):
                return IntVector3.from_tuple(result)
        
        return FloatVector3.from_tuple(result)

    @classmethod
    def from_tuple(cls, xyz:tuple[int, int, int]) -> IntVector3:
        '''
        Instantiate a new Vector from a tuple with integer X, Y, and Z values
        
        Parameters
        ----------
        xyz : `tuple[int, int, int]`
            The tuple with the X, Y, and Z values
        
        Raises
        ------
        `ValueError`
            If the provided items in the tuple are not integers
        '''
        
        return cls(xyz[0], xyz[1], xyz[2])
    
    def normalize(self) -> FloatVector3:
        '''
        Calculate the normal of this vector
        
        Returns
        -------
        vector : `wame.FloatVector3`
            This vector normalized
        '''

        length:float = math.sqrt((self.x ** 2) + (self.y ** 2) + (self.z ** 2))

        if length > 0:
            return FloatVector3(self.x / length, self.y / length, self.z / length)

        return FloatVector3(0.0, 0.0, 0.0)

    def to_numpy(self) -> np.ndarray[np.int32]:
        '''
        Converts this instance of `IntVector3` into a numpy array
        
        Returns
        -------
        array : `numpy.ndarray[numpy.int32]`
            Converted x, y, and z values
        '''

        return np.array([self.x, self.y, self.z], dtype=np.int32)

    def to_tuple(self) -> tuple[int, int, int]:
        '''
        Converts this instance of `IntVector3` into a `tuple`
        
        Returns
        -------
        vector : `tuple[int, int, int]`
            Converted x, y, and z values
        '''

        return (self.x, self.y, self.z)


'''
## Wame Engine
Simple, Pythonic, Intuitive, Pygame Wrapper

#### Version: `0.4.0`
#### Documentation: https://github.com/WilDev-Studios/Wame
'''

from wame.engine import Engine
from wame.pipeline import Pipeline
from wame.scene import Scene
from wame.settings import Settings

from wame import color
from wame import common
from wame import ui
from wame import vector


from __future__ import annotations

from wame.color.rgb import ColorRGB

class RGB:
    '''Common RGB Values'''

    BLACK:ColorRGB = ColorRGB(0, 0, 0)
    '''`RGB: 0, 0, 0`'''

    BLUE:ColorRGB = ColorRGB(0, 0, 255)
    '''`RGB: 0, 0, 255`'''

    CYAN:ColorRGB = ColorRGB(0, 255, 255)
    '''`RGB: 0, 255, 255`'''

    GREEN:ColorRGB = ColorRGB(0, 255, 0)
    '''`RGB: 0, 255, 0`'''

    PINK:ColorRGB = ColorRGB(255, 0, 255)
    '''`RGB: 255, 0, 255`'''

    RED:ColorRGB = ColorRGB(255, 0, 0)
    '''`RGB: 255, 0, 0`'''

    WHITE:ColorRGB = ColorRGB(255, 255, 255)
    '''`RGB: 255, 255, 255`'''

    YELLOW:ColorRGB = ColorRGB(255, 255, 0)
    '''`RGB: 255, 255, 0`'''


from __future__ import annotations

import pygame

KEYS:dict[tuple[int, int], str] = {
    (pygame.K_a, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'A',
    (pygame.K_b, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'B',
    (pygame.K_c, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'C',
    (pygame.K_d, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'D',
    (pygame.K_e, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'E',
    (pygame.K_f, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'F',
    (pygame.K_g, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'G',
    (pygame.K_h, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'H',
    (pygame.K_i, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'I',
    (pygame.K_j, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'J',
    (pygame.K_k, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'K',
    (pygame.K_l, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'L',
    (pygame.K_m, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'M',
    (pygame.K_n, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'N',
    (pygame.K_o, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'O',
    (pygame.K_p, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'P',
    (pygame.K_q, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'Q',
    (pygame.K_r, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'R',
    (pygame.K_s, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'S',
    (pygame.K_t, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'T',
    (pygame.K_u, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'U',
    (pygame.K_v, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'V',
    (pygame.K_w, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'W',
    (pygame.K_x, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'X',
    (pygame.K_y, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'Y',
    (pygame.K_z, pygame.KMOD_SHIFT | pygame.KMOD_CAPS): 'Z',

    (pygame.K_a, pygame.KMOD_NONE): 'a',
    (pygame.K_b, pygame.KMOD_NONE): 'b',
    (pygame.K_c, pygame.KMOD_NONE): 'c',
    (pygame.K_d, pygame.KMOD_NONE): 'd',
    (pygame.K_e, pygame.KMOD_NONE): 'e',
    (pygame.K_f, pygame.KMOD_NONE): 'f',
    (pygame.K_g, pygame.KMOD_NONE): 'g',
    (pygame.K_h, pygame.KMOD_NONE): 'h',
    (pygame.K_i, pygame.KMOD_NONE): 'i',
    (pygame.K_j, pygame.KMOD_NONE): 'j',
    (pygame.K_k, pygame.KMOD_NONE): 'k',
    (pygame.K_l, pygame.KMOD_NONE): 'l',
    (pygame.K_m, pygame.KMOD_NONE): 'm',
    (pygame.K_n, pygame.KMOD_NONE): 'n',
    (pygame.K_o, pygame.KMOD_NONE): 'o',
    (pygame.K_p, pygame.KMOD_NONE): 'p',
    (pygame.K_q, pygame.KMOD_NONE): 'q',
    (pygame.K_r, pygame.KMOD_NONE): 'r',
    (pygame.K_s, pygame.KMOD_NONE): 's',
    (pygame.K_t, pygame.KMOD_NONE): 't',
    (pygame.K_u, pygame.KMOD_NONE): 'u',
    (pygame.K_v, pygame.KMOD_NONE): 'v',
    (pygame.K_w, pygame.KMOD_NONE): 'w',
    (pygame.K_x, pygame.KMOD_NONE): 'x',
    (pygame.K_y, pygame.KMOD_NONE): 'y',
    (pygame.K_z, pygame.KMOD_NONE): 'z',

    (pygame.K_1, pygame.KMOD_SHIFT): '!',
    (pygame.K_2, pygame.KMOD_SHIFT): '@',
    (pygame.K_3, pygame.KMOD_SHIFT): '#',
    (pygame.K_4, pygame.KMOD_SHIFT): '$',
    (pygame.K_5, pygame.KMOD_SHIFT): '%',
    (pygame.K_6, pygame.KMOD_SHIFT): '^',
    (pygame.K_7, pygame.KMOD_SHIFT): '&',
    (pygame.K_8, pygame.KMOD_SHIFT): '*',
    (pygame.K_9, pygame.KMOD_SHIFT): '(',
    (pygame.K_0, pygame.KMOD_SHIFT): ')',

    (pygame.K_1, pygame.KMOD_NONE): '1',
    (pygame.K_2, pygame.KMOD_NONE): '2',
    (pygame.K_3, pygame.KMOD_NONE): '3',
    (pygame.K_4, pygame.KMOD_NONE): '4',
    (pygame.K_5, pygame.KMOD_NONE): '5',
    (pygame.K_6, pygame.KMOD_NONE): '6',
    (pygame.K_7, pygame.KMOD_NONE): '7',
    (pygame.K_8, pygame.KMOD_NONE): '8',
    (pygame.K_9, pygame.KMOD_NONE): '9',
    (pygame.K_0, pygame.KMOD_NONE): '0',

    (pygame.K_BACKQUOTE, pygame.KMOD_SHIFT): '~',
    (pygame.K_MINUS, pygame.KMOD_SHIFT): '_',
    (pygame.K_EQUALS, pygame.KMOD_SHIFT): '+',
    (pygame.K_LEFTBRACKET, pygame.KMOD_SHIFT): '{',
    (pygame.K_RIGHTBRACKET, pygame.KMOD_SHIFT): '}',
    (pygame.K_BACKSLASH, pygame.KMOD_SHIFT): '|',
    (pygame.K_SEMICOLON, pygame.KMOD_SHIFT): ':',
    (pygame.K_QUOTE, pygame.KMOD_SHIFT): '"',
    (pygame.K_COMMA, pygame.KMOD_SHIFT): '<',
    (pygame.K_PERIOD, pygame.KMOD_SHIFT): '>',
    (pygame.K_SLASH, pygame.KMOD_SHIFT): '?',

    (pygame.K_BACKQUOTE, pygame.KMOD_NONE): '`',
    (pygame.K_MINUS, pygame.KMOD_NONE): '-',
    (pygame.K_EQUALS, pygame.KMOD_NONE): '=',
    (pygame.K_LEFTBRACKET, pygame.KMOD_NONE): '[',
    (pygame.K_RIGHTBRACKET, pygame.KMOD_NONE): ']',
    (pygame.K_BACKSLASH, pygame.KMOD_NONE): '\\',
    (pygame.K_SEMICOLON, pygame.KMOD_NONE): ';',
    (pygame.K_QUOTE, pygame.KMOD_NONE): '\'',
    (pygame.K_COMMA, pygame.KMOD_NONE): ',',
    (pygame.K_PERIOD, pygame.KMOD_NONE): '.',
    (pygame.K_SLASH, pygame.KMOD_NONE): '/',
}
'''Mapping of `(KEYCODE, MODIFIERS)` -> `KEY`'''

def is_char(key:int, mods:int) -> bool:
    '''
    Check to see if this key/mods combination equates to a character on the keyboard
    
    Parameters
    ----------
    key : `int`
        The raw `pygame` key code
    mods : `int`
        The raw `pygame` key mods
    
    Returns
    -------
    isChar : `bool`
        If this key/mods combination equates to a character on the keyboard
    '''

    if not KEYS.get((key, mods), None):
        return False

    return True

def is_letter(key:int, mods:int) -> bool:
    '''
    Check to see if this key/mods combination equates to a letter
    
    Parameters
    ----------
    key : `int`
        The raw `pygame` key code
    mods : `int`
        The raw `pygame` key mods
    
    Returns
    -------
    isLetter : `bool`
        If this key/mods combination equates to a letter
    '''

    return is_upper(key, mods) or is_lower(key, mods)

def is_lower(key:int, mods:int) -> bool:
    '''
    Check to see if this key/mods combination equates to a lowercase letter
    
    Parameters
    ----------
    key : `int`
        The raw `pygame` key code
    mods : `int`
        The raw `pygame` key mods
    
    Returns
    -------
    isLower : `bool`
        If this key/mods combination equates to a lowercase letter
    '''

    return KEYS.get((key, mods), None) in [
        'a', 'b', 'c', 'd', 'e', 'f', 'g',
        'h', 'i', 'j', 'k', 'l', 'm', 'n',
        'o', 'p', 'q', 'r', 's', 't', 'u',
        'v', 'w', 'x', 'y', 'z'
    ]

def is_number(key:int, mods:int) -> bool:
    '''
    Check to see if this key/mods combination equates to a number/digit
    
    Parameters
    ----------
    key : `int`
        The raw `pygame` key code
    mods : `int`
        The raw `pygame` key mods
    
    Returns
    -------
    isNumber : `bool`
        If this key/mods combination equates to a number/digit
    '''

    return KEYS.get((key, mods), None) in [
        '1', '2', '3', '4', '5',
        '6', '7', '8', '9', '0'
    ]

def is_symbol(key:int, mods:int) -> bool:
    '''
    Check to see if this key/mods combination equates to a symbol
    
    Parameters
    ----------
    key : `int`
        The raw `pygame` key code
    mods : `int`
        The raw `pygame` key mods
    
    Returns
    -------
    isSymbol : `bool`
        If this key/mods combination equates to a symbol
    '''

    return KEYS.get((key, mods), None) in [
        '~', '`', '!', '@', '#', '$', '%', '^',
        '&', '*', '(', ')', '_', '-', '+', '=',
        '{', '[', '}', ']', ':', ';', '"', '\'',
        '|', '\\', '<', ',', '>', '.', '?', '/'
    ]

def is_upper(key:int, mods:int) -> bool:
    '''
    Check to see if this key/mods combination equates to an uppercase letter
    
    Parameters
    ----------
    key : `int`
        The raw `pygame` key code
    mods : `int`
        The raw `pygame` key mods
    
    Returns
    -------
    isUpper : `bool`
        If this key/mods combination equates to an uppercase letter
    '''

    return KEYS.get((key, mods), None) in [
        'A', 'B', 'C', 'D', 'E', 'F', 'G',
        'H', 'I', 'J', 'K', 'L', 'M', 'N',
        'O', 'P', 'Q', 'R', 'S', 'T', 'U',
        'V', 'W', 'X', 'Y', 'Z'
    ]


'''
## Vector Module - Wame Engine
Contains all necessary vector objects for 2D, 3D, and 4D positional, sizeable, rotational, and color fields
'''

from wame.vector.xy import FloatVector2, IntVector2
from wame.vector.xyz import FloatVector3, IntVector3


from __future__ import annotations

from typing import Callable

from wame.color.rgb import ColorRGBA
from wame.common.keys import KEYS
from wame.vector.xy import FloatVector2, IntVector2
from wame.ui.frame import Frame
from wame.ui.renderable import Renderable
from wame.ui.text import Text

from OpenGL.GLU import *
from OpenGL.GL import *

import pygame
import wame

class CheckboxInput(Renderable):
    '''UI Checkbox Input'''

    def __init__(self, parent:'Frame', checkedColor:ColorRGBA, uncheckedColor:ColorRGBA, *, default:bool=False, y_flipped:bool=False) -> None:
        '''
        Create a UI checkbox input
        
        Parameters
        ----------
        parent : `wame.ui.frame.Frame`
            The parent of this input
        checkedColor : `wame.color.rgb.ColorRGBA`
            The color of the box when checked (`True`/active)
        uncheckedColor : `wame.color.rgb.ColorRGBA`
            The color of the box when unchecked (`False`/inactive)
        default : `bool`
            The original state of the input before any interaction - Default `False`
        y_flipped : `bool`
            If it should be rendered with the Y-axis flipped - May be necessary depending on your OpenGL setup
        '''

        super().__init__(parent._engine)

        parent.add_child(self)
        self._parent = parent

        self._children:list[Renderable] = []

        self._checked_color:ColorRGBA = checkedColor if isinstance(checkedColor, ColorRGBA) else ColorRGBA.from_tuple(checkedColor)
        self._unchecked_color:ColorRGBA = uncheckedColor if isinstance(uncheckedColor, ColorRGBA) else ColorRGBA.from_tuple(uncheckedColor)
        self.state:bool = default
        '''The current state of the checkbox (`True`/`False`)'''

        self._border_color:ColorRGBA = None
        self._border_width:int = None

        self._flipped:bool = y_flipped

        self._callback:Callable[[bool], None] = None

    def check_click(self, position:IntVector2) -> None:
        '''
        Check if the location in which a mouse click is able to interact with this input
        
        Parameters
        ----------
        position : `wame.vector.xy.IntVector2`
            The X, Y vector/position of the mouse
        '''

        if not self.rect.collidepoint(position.to_tuple()):
            return
        
        self.state = not self.state

        if self._callback:
            self._callback()

    def render(self) -> None:
        '''
        Render this input and it's children to the screen
        '''
        
        if not self.position or not self.size:
            error:str = "The frame must have its size and position set before rendering"
            raise ValueError(error)
    
        if self._engine._pipeline == wame.Pipeline.PYGAME:
            pygame.draw.rect(
                self._engine.screen, self._checked_color.to_tuple() if self.state else self._unchecked_color.to_tuple(), self.rect
            )
        elif self._engine._pipeline == wame.Pipeline.OPENGL:
            posY:int = self.position.y
            posWidth:int = self.position.x + self.size.x
            posHeight:int = self.position.y + self.size.y

            if not self._flipped:
                posY = self._engine.screen.get_height() - (self.position.y + self.size.y)

            glPushMatrix()

            glBegin(GL_QUADS)
            if self.state:
                glColor4f(*self._checked_color.to_tuple())
            else:
                glColor4f(*self._unchecked_color.to_tuple())
            if self._flipped:
                glVertex2f(self.position.x, posHeight)
                glVertex2f(posWidth, posHeight)
                glVertex2f(posWidth, posY)
                glVertex2f(self.position.x, posY)
            else:
                glVertex2f(self.position.x, posY)
                glVertex2f(posWidth, posY)
                glVertex2f(posWidth, posHeight)
                glVertex2f(self.position.x, posHeight)
            glEnd()

            glPopMatrix()
        
        if self._border_color and self._border_width >= 1:
            # Lines are straight, no point in antialiasing them

            if self._engine._pipeline == wame.Pipeline.PYGAME:
                for index in range(self._border_width):
                    pygame.draw.lines(self._engine.screen, self._border_color.to_tuple(), True, [
                        (self.position.x + index, self.position.y + index), (self.position.x + self.size.x - index, self.position.y + index),
                        (self.position.x + self.size.x - index, self.position.y + self.size.y - index), (self.position.x + index, self.position.y + self.size.y - index)
                    ])
            elif self._engine._pipeline == wame.Pipeline.OPENGL:
                posX:int = self.position.x
                posY:int = self.position.y
                posWidth:int = self.position.x + self.size.x
                posHeight:int = self.position.y + self.size.y

                if not self._flipped:
                    posY = self._engine.screen.get_height() - (self.position.y + self.size.y)

                for index in range(self._border_width):
                    glPushMatrix()

                    glBegin(GL_LINES)
                    glColor4f(*self._border_color.to_tuple())
                    glVertex2f(posX + index, posY + index)
                    glVertex2f(posWidth - index, posY + index)

                    glVertex2f(posWidth - index, posY + index)
                    glVertex2f(posWidth - index, posHeight - index)

                    glVertex2f(posWidth - index, posHeight - index)
                    glVertex2f(posX + index, posHeight - index)

                    glVertex2f(posX + index, posHeight - index)
                    glVertex2f(posX + index, posY + index)
                    glEnd()

                    glPopMatrix()

        for child in self._children:
            child.ask_render()
    
    def set_border(self, color:ColorRGBA, width:int) -> None:
        self._border_color = color if isinstance(color, ColorRGBA) else ColorRGBA.from_tuple(color)
        self._border_width = width

    def set_callback(self, func:Callable[[bool], None]) -> None:
        '''
        Set the callback method for when this input is edited/activated/deactivated
        
        Parameters
        ----------
        func : `typing.Callable[[bool], None]`
            The callback method to call - Takes the state of the input (`True`/`False`) as the only parameter
        '''

        self._callback = func

    def set_pixel_position(self, position:IntVector2) -> None:
        position = position if isinstance(position, IntVector2) else IntVector2.from_tuple(position)

        if self._parent:
            position.x += self._parent.position.x
            position.y += self._parent.position.y
        
        self.position = position

    def set_pixel_size(self, size:IntVector2) -> None:
        self.size = size if isinstance(size, IntVector2) else IntVector2.from_tuple(size)

    def set_scaled_position(self, position:FloatVector2) -> None:
        position = position if isinstance(position, FloatVector2) else FloatVector2.from_tuple(position)

        if position.x > 1 or position.x < 0 or position.y > 1 or position.y < 0:
            error:str = "Scaled position X, Y values must be between 0 and 1"
            raise ValueError(error)
        
        newPosition:IntVector2 = IntVector2(0, 0)

        if not self._parent:
            newPosition.x = int(self._engine.screen.get_width() * position.x)
            newPosition.y = int(self._engine.screen.get_height() * position.y)
        else:
            newPosition.x = int(self._parent.position.x + (self._parent.size.x * position.x))
            newPosition.y = int(self._parent.position.y + (self._parent.size.y * position.y))

        self.position = newPosition
    
    def set_scaled_size(self, size:IntVector2) -> None:
        size = size if isinstance(size, FloatVector2) else FloatVector2.from_tuple(size)

        if size.x > 1 or size.x < 0 or size.y > 1 or size.y < 0:
            error:str = "Scaled size X, Y values must be between 0 and 1"
            raise ValueError(error)
    
        newSize:IntVector2 = IntVector2(0, 0)

        if not self._parent:
            newSize.x = int(self._engine.screen.get_width() * size.x)
            newSize.y = int(self._engine.screen.get_height() * size.y)
        else:
            newSize.x = int(self._parent.size.x * size.x)
            newSize.y = int(self._parent.size.y * size.y)

        self.size = newSize

class TextInput(Renderable):
    '''UI Text Input'''

    def __init__(self, parent:'Frame', textColor:ColorRGBA, font:pygame.font.Font, *, default:str=None, yFlipped:bool=False) -> None:
        '''
        Create a UI text input
        
        Parameters
        ----------
        parent : `wame.ui.frame.Frame`
            The parent of this input
        textColor : `wame.color.rgb.ColorRGBA`
            The color of the text inside of the input
        font : `pygame.font.Font`
            The font of the text inside of the input
        default : `str`
            The original text to show in the input before any interaction - Default `None`
        yFlipped : `bool`
            If it should be rendered with the Y-axis flipped - May be necessary depending on your OpenGL setup
        '''

        super().__init__(parent._engine)

        parent.add_child(self)
        self._parent = parent

        self._children:list[Renderable] = []

        self._color:ColorRGBA = None

        self._border_color:ColorRGBA = None
        self._border_width:int = None

        self._flipped:bool = yFlipped
        
        self.text:Text = Text(self, default if default else "", textColor, font, yFlipped)
        '''The internal text object of this input'''

        self._active:bool = False
        self._callback:Callable[[None], None] = None

    def check_click(self, position:IntVector2) -> None:
        '''
        Check if the location in which a mouse click is able to interact with this input
        
        Parameters
        ----------
        position : `wame.vector.xy.IntVector2`
            The X, Y vector/position of the mouse
        '''

        if self.rect.collidepoint(position.to_tuple()):
            active:bool = self._active
            self._active = True

            if not active and self._callback:
                self._callback()
        else:
            active:bool = self._active
            self._active = False
            
            if active and self._callback:
                self._callback()
    
    def check_key(self, key:int, mods:int, predicate:Callable[[int, int], bool]=None) -> None:
        '''
        Check if a key/mod combination is able to interact with this input (if active)
        
        Parameters
        ----------
        key : `int`
            The raw `pygame` keycode input
        mods : `int`
            The raw `pygame` key modifications
        predicate : `typing.Callable[[int, int], bool]`
            A function object that should be used to filter out whether the input should or should not be accepted - Default `None` (all will be entered)
        '''

        if not self._active:
            return
        
        if predicate:
            if not predicate(key, mods):
                return
            
        strKey:str = KEYS.get((key, mods), None)

        if not strKey:
            if key == pygame.K_BACKSPACE and len(self.text.raw_text) >= 1:
                text:str = self.text.raw_text[:-1]

                self.text.set_text(text)

                if self._callback:
                    self._callback()

            return
        
        self.text.set_text(self.text.raw_text + strKey)

        if self._callback:
            self._callback()

    def render(self) -> None:
        '''
        Render this input and it's children to the screen
        '''
        
        if not self.position or not self.size:
            error:str = "The frame must have its size and position set before rendering"
            raise ValueError(error)
    
        if self._color:
            if self._engine._pipeline == wame.Pipeline.PYGAME:
                pygame.draw.rect(
                    self._engine.screen, self._color.to_tuple(), self.rect
                )
            elif self._engine._pipeline == wame.Pipeline.OPENGL:
                posY:int = self.position.y
                posWidth:int = self.position.x + self.size.x
                posHeight:int = self.position.y + self.size.y

                if not self._flipped:
                    posY = self._engine.screen.get_height() - (self.position.y + self.size.y)

                glPushMatrix()

                glBegin(GL_QUADS)
                glColor4f(*self._color.to_tuple())
                if self._flipped:
                    glVertex2f(self.position.x, posHeight)
                    glVertex2f(posWidth, posHeight)
                    glVertex2f(posWidth, posY)
                    glVertex2f(self.position.x, posY)
                else:
                    glVertex2f(self.position.x, posY)
                    glVertex2f(posWidth, posY)
                    glVertex2f(posWidth, posHeight)
                    glVertex2f(self.position.x, posHeight)
                glEnd()

                glPopMatrix()
        
        if self._border_color and self._border_width >= 1:
            # Lines are straight, no point in antialiasing them

            if self._engine._pipeline == wame.Pipeline.PYGAME:
                for index in range(self._border_width):
                    pygame.draw.lines(self._engine.screen, self._border_color.to_tuple(), True, [
                        (self.position.x + index, self.position.y + index), (self.position.x + self.size.x - index, self.position.y + index),
                        (self.position.x + self.size.x - index, self.position.y + self.size.y - index), (self.position.x + index, self.position.y + self.size.y - index)
                    ])
            elif self._engine._pipeline == wame.Pipeline.OPENGL:
                posX:int = self.position.x
                posY:int = self.position.y
                posWidth:int = self.position.x + self.size.x
                posHeight:int = self.position.y + self.size.y

                if not self._flipped:
                    posY = self._engine.screen.get_height() - (self.position.y + self.size.y)

                for index in range(self._border_width):
                    glPushMatrix()

                    glBegin(GL_LINES)
                    glColor4f(*self._border_color.to_tuple())
                    glVertex2f(posX + index, posY + index)
                    glVertex2f(posWidth - index, posY + index)

                    glVertex2f(posWidth - index, posY + index)
                    glVertex2f(posWidth - index, posHeight - index)

                    glVertex2f(posWidth - index, posHeight - index)
                    glVertex2f(posX + index, posHeight - index)

                    glVertex2f(posX + index, posHeight - index)
                    glVertex2f(posX + index, posY + index)
                    glEnd()

                    glPopMatrix()

        for child in self._children:
            child.ask_render()
    
    def set_border(self, color:ColorRGBA, width:int) -> None:
        self._border_color = color if isinstance(color, ColorRGBA) else ColorRGBA.from_tuple(color)
        self._border_width = width

    def set_callback(self, func:Callable[[None], None]) -> None:
        '''
        Set the callback method for when this input is edited/activated/deactivated
        
        Parameters
        ----------
        func : `typing.Callable[[None], None]`
            The callback method to call
        '''

        self._callback = func

    def set_color(self, color:ColorRGBA) -> None:
        self._color = color if isinstance(color, ColorRGBA) else ColorRGBA.from_tuple(color)

    def set_pixel_position(self, position:IntVector2) -> None:
        position = position if isinstance(position, IntVector2) else IntVector2.from_tuple(position)

        if self._parent:
            position.x += self._parent.position.x
            position.y += self._parent.position.y
        
        self.position = position

        if self.size:
            self.text.set_pixel_position((
                10, (self.size.y // 2) - (self.text.rect.height // 2)
            ))

    def set_pixel_size(self, size:IntVector2) -> None:
        self.size = size if isinstance(size, IntVector2) else IntVector2.from_tuple(size)

        if self.position:
            self.text.set_pixel_position((
                10, (self.size.y // 2) - (self.text.text.get_height() // 2)
            ))

    def set_scaled_position(self, position:FloatVector2) -> None:
        position = position if isinstance(position, FloatVector2) else FloatVector2.from_tuple(position)

        if position.x > 1 or position.x < 0 or position.y > 1 or position.y < 0:
            error:str = "Scaled position X, Y values must be between 0 and 1"
            raise ValueError(error)
        
        newPosition:IntVector2 = IntVector2(0, 0)

        if not self._parent:
            newPosition.x = int(self._engine.screen.get_width() * position.x)
            newPosition.y = int(self._engine.screen.get_height() * position.y)
        else:
            newPosition.x = int(self._parent.position.x + (self._parent.size.x * position.x))
            newPosition.y = int(self._parent.position.y + (self._parent.size.y * position.y))

        self.position = newPosition

        if self.size:
            self.text.set_pixel_position((
                10, (self.size.y // 2) - (self.text.rect.height // 2)
            ))
    
    def set_scaled_size(self, size:IntVector2) -> None:
        size = size if isinstance(size, FloatVector2) else FloatVector2.from_tuple(size)

        if size.x > 1 or size.x < 0 or size.y > 1 or size.y < 0:
            error:str = "Scaled size X, Y values must be between 0 and 1"
            raise ValueError(error)
    
        newSize:IntVector2 = IntVector2(0, 0)

        if not self._parent:
            newSize.x = int(self._engine.screen.get_width() * size.x)
            newSize.y = int(self._engine.screen.get_height() * size.y)
        else:
            newSize.x = int(self._parent.size.x * size.x)
            newSize.y = int(self._parent.size.y * size.y)

        self.size = newSize

        self.text.set_pixel_position((
                10, (self.size.y // 2) - (self.text.rect.height // 2)
            ))


from __future__ import annotations

from typing import Union

import numpy as np

class FloatVector2:
    '''Vector with 2 Float Values: X and Y'''

    def __init__(self, x:float, y:float) -> None:
        '''
        Instantiate a new Vector with float X and Y values
        
        Parameters
        ----------
        x : `float`
            The X value
        y : `float`
            The Y value
        
        Raises
        ------
        `ValueError`
            If any provided arguments are not floats
        '''
        
        if isinstance(x, int):
            x = float(x)
        
        if isinstance(y, int):
            y = float(y)

        self.x:float = x
        self.y:float = y
    
    def __add__(self, other:Union[FloatVector2, IntVector2, np.ndarray]) -> FloatVector2:
        if isinstance(other, FloatVector2) or isinstance(other, IntVector2):
            return FloatVector2(self.x + other.x, self.y + other.y)
        
        if isinstance(other, np.ndarray):
            return FloatVector2(self.x + other[0], self.y + other[1])
        
        raise TypeError(f"Unsupported addition with type {type(other)}")
    
    def __eq__(self, other:Union[FloatVector2, IntVector2, np.ndarray]) -> bool:
        if isinstance(other, FloatVector2) or isinstance(other, IntVector2):
            return self.to_tuple() == other.to_tuple()
        
        if isinstance(other, np.ndarray):
            return self.to_tuple() == (other[0], other[1])

    def __hash__(self) -> int:
        return hash(self.to_tuple())

    def __sub__(self, other:Union[FloatVector2, IntVector2, np.ndarray]) -> FloatVector2:
        if isinstance(other, FloatVector2) or isinstance(other, IntVector2):
            return FloatVector2(self.x - other.x, self.y - other.y)
        
        if isinstance(other, np.ndarray):
            return FloatVector2(self.x - other[0], self.y - other[1])
        
        raise TypeError(f"Unsupported subtraction with type {type(other)}")

    def __repr__(self) -> str:
        return f"X: {self.x}, Y: {self.y}"
    
    @classmethod
    def from_tuple(cls, xy:tuple[float, float]) -> FloatVector2:
        '''
        Instantiate a new Vector from a tuple with float X and Y values
        
        Parameters
        ----------
        xy : `tuple[float, float]`
            The tuple with the X and Y values
        
        Raises
        ------
        `ValueError`
            If the provided items in the tuple are not floats
        '''
        
        return cls(xy[0], xy[1])
    
    def to_numpy(self) -> np.ndarray[np.int32]:
        '''
        Convers this instance of `FloatVector2` into a numpy array
        
        Returns
        -------
        array : `numpy.ndarray[numpy.float32]`
            Converted x and Y values
        '''

        return np.array([self.x, self.y], dtype=np.float32)

    def to_tuple(self) -> tuple[float, float]:
        '''
        Converts this instance of `FloatVector2` into a `tuple`
        
        Returns
        -------
        vector : `tuple[float, float]`
            Converted x and y values
        '''

        return (self.x, self.y)

class IntVector2:
    '''Vector with 2 Integer Values: X and Y'''

    def __init__(self, x:int, y:int) -> None:
        '''
        Instantiate a new Vector with integer X and Y values
        
        Parameters
        ----------
        x : `int`
            The X value
        y : `int`
            The Y value
        
        Raises
        ------
        `ValueError`
            If any provided arguments are not integers
        '''
        
        if not isinstance(x, int):
            error:str = f"Value of X ({x}) is not an integer"
            raise ValueError(error)
    
        if not isinstance(y, int):
            error:str = f"Value of Y ({y}) is not an integer"
            raise ValueError(error)

        self.x:int = x
        self.y:int = y
    
    def __add__(self, other:Union[IntVector2, np.ndarray]) -> IntVector2:
        if isinstance(other, IntVector2):
            return IntVector2(self.x + other.x, self.y + other.y)
        
        if isinstance(other, np.ndarray):
            return IntVector2(self.x + other[0], self.y + other[1])
        
        raise TypeError(f"Unsupported addition with type {type(other)}")
    
    def __eq__(self, other:Union[FloatVector2, IntVector2, np.ndarray]) -> bool:
        if isinstance(other, FloatVector2) or isinstance(other, IntVector2):
            return self.to_tuple() == other.to_tuple()
        
        if isinstance(other, np.ndarray):
            return self.to_tuple() == (other[0], other[1])

    def __hash__(self) -> int:
        return hash(self.to_tuple())

    def __sub__(self, other:Union[IntVector2, np.ndarray]) -> IntVector2:
        if isinstance(other, IntVector2):
            return IntVector2(self.x - other.x, self.y - other.y)
        
        if isinstance(other, np.ndarray):
            return IntVector2(self.x - other[0], self.y - other[1])
        
        raise TypeError(f"Unsupported subtraction with type {type(other)}")

    def __repr__(self) -> str:
        return f"X: {self.x}, Y: {self.y}"

    @classmethod
    def from_tuple(cls, xy:tuple[int, int]) -> IntVector2:
        '''
        Instantiate a new Vector from a tuple with integer X and Y values
        
        Parameters
        ----------
        xy : `tuple[int, int]`
            The tuple with the X and Y values
        
        Raises
        ------
        `ValueError`
            If the provided items in the tuple are not integers
        '''
        
        return cls(xy[0], xy[1])
    
    def to_numpy(self) -> np.ndarray[np.int32]:
        '''
        Converts this instance of `IntVector2` into a numpy array
        
        Returns
        -------
        array : `numpy.ndarray[numpy.int32]`
            Converted x and y values
        '''

        return np.array([self.x, self.y], dtype=np.int32)

    def to_tuple(self) -> tuple[int, int]:
        '''
        Converts this instance of `IntVector2` into a `tuple`
        
        Returns
        -------
        vector : `tuple[int, int]`
            Converted x and y values
        '''

        return (self.x, self.y)


'''
## Common Module - Wame Engine
Contains all objects and functionality that might be useful to developers
'''

from wame.common.keys import KEYS, is_char, is_letter, is_lower, is_number, is_symbol, is_upper


from __future__ import annotations

from typing import Callable

from wame.ui.frame import Frame
from wame.ui.renderable import Renderable
from wame.vector.xy import IntVector2, FloatVector2

from OpenGL.GL import *

class Button(Renderable):
    '''UI Button Object'''

    def __init__(self, parent:Frame) -> None:
        super().__init__(parent._engine)

        parent.add_child(self)
        self._parent:Frame = parent

        self._click_callback:Callable = None

        self._hovering:bool = False
        self._hover_callback:Callable = None
        self._unhover_callback:Callable = None

    def check_click(self, position:IntVector2) -> None:
        if not self.rect.collidepoint(position.to_tuple()):
            return
        
        if not self._click_callback:
            return
        
        self._click_callback()
    
    def check_hover(self, position:IntVector2) -> None:
        if self.rect.collidepoint(position.to_tuple()):
            if not self._hovering:
                self._hovering = True

                if self._hover_callback:
                    self._hover_callback()
        else:
            if self._hovering:
                self._hovering = False

                if self._unhover_callback:
                    self._unhover_callback()

    def render(self) -> None:
        for child in self._children:
            child.ask_render()

    def set_click_callback(self, func:Callable[[], None]) -> None:
        self._click_callback = func
    
    def set_hover_callback(self, func:Callable[[], None]) -> None:
        self._hover_callback = func

    def set_pixel_position(self, position:IntVector2) -> None:
        position = position if isinstance(position, IntVector2) else IntVector2.from_tuple(position)

        if self._parent:
            position.x += self._parent.position.x
            position.y += self._parent.position.y
        
        self.position = position

    def set_pixel_size(self, size:IntVector2) -> None:
        self.size = size if isinstance(size, IntVector2) else IntVector2.from_tuple(size)

    def set_scaled_position(self, position:FloatVector2) -> None:
        position = position if isinstance(position, FloatVector2) else FloatVector2.from_tuple(position)

        if position.x > 1 or position.x < 0 or position.y > 1 or position.y < 0:
            error:str = "Scaled position X, Y values must be between 0 and 1"
            raise ValueError(error)

        self.position = IntVector2(
            int(self._parent.position.x + (self._parent.size.x * position.x)),
            int(self._parent.position.y + (self._parent.size.y * position.y))
        )
    
    def set_scaled_size(self, size:IntVector2) -> None:
        size = size if isinstance(size, FloatVector2) else FloatVector2.from_tuple(size)

        if size.x > 1 or size.x < 0 or size.y > 1 or size.y < 0:
            error:str = "Scaled size X, Y values must be between 0 and 1"
            raise ValueError(error)

        self.size = IntVector2(
            int(self._parent.size.x * size.x),
            int(self._parent.size.y * size.y)
        )
    
    def set_unhover_callback(self, func:Callable) -> None:
        self._unhover_callback = func


from __future__ import annotations

from wame.color.rgb import ColorRGBA
from wame.ui.frame import Frame
from wame.ui.renderable import Renderable
from wame.vector.xy import IntVector2, FloatVector2

from wame.pipeline import Pipeline

from OpenGL.GL import *

import pygame

class Text(Renderable):
    '''UI Text Object'''

    def __init__(self, parent:Frame, text:str, color:ColorRGBA, font:pygame.font.Font, y_flipped:bool=False) -> None:
        super().__init__(parent._engine)

        color = color if isinstance(color, ColorRGBA) else ColorRGBA.from_tuple(color)

        parent.add_child(self)
        self._parent:Frame = parent

        self._color:ColorRGBA = color
        self._colors:dict[str, ColorRGBA] = {
            "default": color
        }

        self._font:pygame.font.Font = font
        self._fonts:dict[str, pygame.font.Font] = {
            "default": font
        }

        self.raw_text:str = text
        self.text:pygame.Surface = font.render(text, self._engine.settings.antialiasing, color.to_tuple())
        self._gl_texture_id = None
        self._flipped:bool = y_flipped

        self._render_text()

    def _render_text(self) -> None:
        self.text = self._font.render(self.raw_text, self._engine.settings.antialiasing, self._color.to_tuple())

        if self._engine._pipeline == Pipeline.OPENGL:
            if self._gl_texture_id:
                glDeleteTextures([self._gl_texture_id])

            textData = pygame.image.tostring(self.text, "RGBA", True)
            width, height = self.text.get_size()

            self._gl_texture_id = glGenTextures(1)
            glBindTexture(GL_TEXTURE_2D, self._gl_texture_id)

            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)

            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, textData)
            glBindTexture(GL_TEXTURE_2D, 0)

    def add_color(self, name:str, color:ColorRGBA, overwrite:bool=False) -> None:
        if name in self._colors and not overwrite:
            error:str = f"Color with name {name} is already registered as a color"
            raise ValueError(error)

        self._colors[name] = color if isinstance(color, ColorRGBA) else ColorRGBA.from_tuple(color)

    def add_font(self, name:str, font:pygame.font.Font) -> None:
        if name in self._fonts:
            error:str = f"Font with name {name} is already registered as a font"
            raise ValueError(error)
        
        self._fonts[name] = font

    @property
    def rect(self) -> pygame.Rect:
        return pygame.Rect(self.position.to_tuple(), self.text.get_size())

    def render(self) -> None:
        if self._engine._pipeline == Pipeline.PYGAME:
            self._engine.screen.blit(self.text, self.position.to_tuple())
        elif self._engine._pipeline == Pipeline.OPENGL:
            if not self._gl_texture_id:
                return
            
            width:int = self.text.get_width()
            height:int = self.text.get_height()

            posX:int = self.position.x
            posY:int = self.position.y

            if not self._flipped:
                posY = self._engine.screen.get_height() - (self.position.y + self.text.get_height())

            glPushMatrix()

            glEnable(GL_TEXTURE_2D)
            glBindTexture(GL_TEXTURE_2D, self._gl_texture_id)

            glEnable(GL_BLEND)
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

            if self._engine.settings.antialiasing:
                glEnable(GL_MULTISAMPLE)

            glColor4f(1.0, 1.0, 1.0, 1.0)

            glBegin(GL_QUADS)
            if self._flipped:
                glTexCoord2f(0, 1)
                glVertex2f(posX, posY)
                glTexCoord2f(1, 1)
                glVertex2f(posX + width, posY)
                glTexCoord2f(1, 0)
                glVertex2f(posX + width, posY + height)
                glTexCoord2f(0, 0)
                glVertex2f(posX, posY + height)
            else:
                glTexCoord2f(0, 0)
                glVertex2f(posX, posY)
                glTexCoord2f(1, 0)
                glVertex2f(posX + width, posY)
                glTexCoord2f(1, 1)
                glVertex2f(posX + width, posY + height)
                glTexCoord2f(0, 1)
                glVertex2f(posX, posY + height)
            glEnd()

            glBindTexture(GL_TEXTURE_2D, 0)
            glDisable(GL_BLEND)
            glDisable(GL_TEXTURE_2D)

            if self._engine.settings.antialiasing:
                glDisable(GL_MULTISAMPLE)

            glPopMatrix()

        for child in self._children:
            child.ask_render()

    def set_color(self, name:str) -> None:
        if name not in self._colors:
            error:str = f"Color with name {name} is not registered as a color"
            raise ValueError(error)
        
        self._color = self._colors[name]
        self._render_text()
    
    def set_font(self, name:str) -> None:
        if name not in self._fonts:
            error:str = f"Font with name {name} is not registered as a font"
            raise ValueError(error)
        
        self._font = self._fonts[name]
        self._render_text()
    
    def set_pixel_position(self, position:IntVector2) -> None:
        position = position if isinstance(position, IntVector2) else IntVector2.from_tuple(position)
        position.x += self._parent.position.x
        position.y += self._parent.position.y

        self.position = position
    
    def set_scaled_position(self, position:FloatVector2) -> None:
        position = position if isinstance(position, FloatVector2) else FloatVector2.from_tuple(position)

        if position.x > 1 or position.x < 0 or position.y > 1 or position.y < 0:
            error:str = "Scaled position X, Y values must be between 0 and 1"
            raise ValueError(error)
        
        newPosition:IntVector2 = IntVector2(0, 0)
        newPosition.x = int(self._parent.position.x + (self._parent.size.x * position.x))
        newPosition.y = int(self._parent.position.y + (self._parent.size.y * position.y))

        self.position = newPosition
    
    def set_text(self, text:str) -> None:
        self.raw_text = text
        self._render_text()